@@ -39,16 +39,18 @@
 #include "ev-print-job.h"
 #include "ev-document-thumbnails.h"
 #include "ev-document-links.h"
+#include "ev-document-types.h"
 #include "ev-document-find.h"
 #include "ev-document-security.h"
+#include "ev-job-queue.h"
 #include "eggfindbar.h"
-
-#include "pdf-document.h"
-#include "pixbuf-document.h"
-#include "ps-document.h"
+#include "egg-recent-view-gtk.h"
+#include "egg-recent-view.h"
+#include "egg-recent-model.h"
 
 #include <glib/gi18n.h>
 #include <gtk/gtk.h>
+#include <gnome.h>
 
 #include <libgnomevfs/gnome-vfs-mime-utils.h>
 #include <libgnomevfs/gnome-vfs-uri.h>
@@ -90,7 +92,6 @@ struct _EvWindowPrivate {
 	GtkWidget *toolbar;
 	GtkWidget *hpaned;
 	GtkWidget *sidebar;
-	GtkWidget *thumbs_sidebar;
 	GtkWidget *find_bar;
 	GtkWidget *scrolled_window;
 	GtkWidget *view;
@@ -106,6 +107,7 @@ struct _EvWindowPrivate {
 	char *uri;
 
 	EvDocument *document;
+	EvPageCache *page_cache;
 
 	EvWindowPageMode page_mode;
 	/* These members are used temporarily when in PAGE_MODE_PASSWORD */
@@ -117,9 +119,13 @@ struct _EvWindowPrivate {
 	gboolean fullscreen_mode;
 	EvSizingMode sizing_mode;
 	GSource *fullscreen_timeout_source;
+
+	/* recent file stuff */
+	EggRecentModel *recent_model;
+	EggRecentViewGtk *recent_view;
 };
 
-static GtkTargetEntry ev_drop_types[] = {
+static const GtkTargetEntry ev_drop_types[] = {
 	{ "text/uri-list", 0, 0 }
 };
 
@@ -134,7 +140,9 @@ static GtkTargetEntry ev_drop_types[] = {
 #define GCONF_CHROME_STATUSBAR	"/apps/evince/show_statusbar"
 
 #define GCONF_SIDEBAR_SIZE      "/apps/evince/sidebar_size"
+
 #define SIDEBAR_DEFAULT_SIZE    132
+#define VIEW_SPACING		10
 
 static void     ev_window_update_fullscreen_popup (EvWindow         *window);
 static void     ev_window_sidebar_visibility_changed_cb (EvSidebar *ev_sidebar, GParamSpec *pspec,
@@ -147,6 +155,9 @@ static gboolean start_loading_document            (EvWindow         *ev_window,
 static void     ev_window_set_sizing_mode         (EvWindow         *ev_window,
 						   EvSizingMode      sizing_mode);
 
+static void 	ev_window_add_recent (EvWindow *window, const char *filename);
+static void	ev_window_fullscreen (EvWindow *window);
+
 G_DEFINE_TYPE (EvWindow, ev_window, GTK_TYPE_WINDOW)
 
 static void
@@ -162,50 +173,54 @@ set_action_sensitive (EvWindow   *ev_window,
 static void
 update_action_sensitivity (EvWindow *ev_window)
 {
+	EvView *view;
 	EvDocument *document;
 	EvWindowPageMode page_mode;
-	EvView *view;
-
-	document = ev_window->priv->document;
-	page_mode = ev_window->priv->page_mode;
+	gboolean sensitive, has_pages = FALSE, has_document;
+	int n_pages = 0, page = -1;
 
 	view = EV_VIEW (ev_window->priv->view);
+	document = ev_window->priv->document;
+	page_mode = ev_window->priv->page_mode;
+	has_document = document != NULL;
+	if (has_document && ev_window->priv->page_cache) {
+		page = ev_page_cache_get_current_page (ev_window->priv->page_cache);
+		n_pages = ev_page_cache_get_n_pages (ev_window->priv->page_cache);
+		has_pages = has_document && n_pages > 0;
+	}
 
 	/* File menu */
 	/* "FileOpen": always sensitive */
-	set_action_sensitive (ev_window, "FileSaveAs", document!=NULL);
-	set_action_sensitive (ev_window, "FilePrint", document!=NULL);
+	set_action_sensitive (ev_window, "FileSaveAs", has_document);
+	set_action_sensitive (ev_window, "FilePrint", has_pages);
 	/* "FileCloseWindow": always sensitive */
 
         /* Edit menu */
-	set_action_sensitive (ev_window, "EditCopy", document!=NULL);
-	set_action_sensitive (ev_window, "EditSelectAll", document!=NULL);
 
-	if (document)
-		set_action_sensitive (ev_window, "EditFind", EV_IS_DOCUMENT_FIND (document));
-	else
-		set_action_sensitive (ev_window, "EditFind", FALSE);
+	sensitive = has_pages && ev_document_can_get_text (document);
+	set_action_sensitive (ev_window, "EditCopy", sensitive);
+	set_action_sensitive (ev_window, "EditSelectAll", sensitive);
+	set_action_sensitive (ev_window, "EditFind",
+			      has_pages && EV_IS_DOCUMENT_FIND (document));
+	set_action_sensitive (ev_window, "EditFindNext",
+			      ev_view_can_find_next (view));
 
         /* View menu */
-	set_action_sensitive (ev_window, "ViewZoomIn", document!=NULL);
-	set_action_sensitive (ev_window, "ViewZoomOut", document!=NULL);
-	set_action_sensitive (ev_window, "ViewNormalSize", document!=NULL);
-	set_action_sensitive (ev_window, "ViewBestFit", document!=NULL);
-	set_action_sensitive (ev_window, "ViewPageWidth", document!=NULL);
-	set_action_sensitive (ev_window, "ViewReload", document!=NULL);
+	set_action_sensitive (ev_window, "ViewZoomIn",
+			      has_pages && ev_view_can_zoom_in (view));
+	set_action_sensitive (ev_window, "ViewZoomOut",
+			      has_pages && ev_view_can_zoom_out (view));
+	set_action_sensitive (ev_window, "ViewNormalSize", has_pages);
+	set_action_sensitive (ev_window, "ViewBestFit", has_pages);
+	set_action_sensitive (ev_window, "ViewPageWidth", has_pages);
+	set_action_sensitive (ev_window, "ViewReload", has_pages);
 
         /* Go menu */
 	if (document) {
-		int n_pages;
-		int page;
-
-		page = ev_view_get_page (EV_VIEW (ev_window->priv->view));
-		n_pages = ev_document_get_n_pages (document);
-
-		set_action_sensitive (ev_window, "GoPreviousPage", page > 1);
-		set_action_sensitive (ev_window, "GoNextPage", page < n_pages);
-		set_action_sensitive (ev_window, "GoFirstPage", page > 1);
-		set_action_sensitive (ev_window, "GoLastPage", page < n_pages);
+		set_action_sensitive (ev_window, "GoPreviousPage", page > 0);
+		set_action_sensitive (ev_window, "GoNextPage", page < n_pages - 1);
+		set_action_sensitive (ev_window, "GoFirstPage", page > 0);
+		set_action_sensitive (ev_window, "GoLastPage", page < n_pages - 1);
 	} else {
   		set_action_sensitive (ev_window, "GoFirstPage", FALSE);
 		set_action_sensitive (ev_window, "GoPreviousPage", FALSE);
@@ -226,7 +241,7 @@ update_action_sensitivity (EvWindow *ev_window)
 	/* "HelpAbout": always sensitive */
 
 	/* Toolbar-specific actions: */
-	set_action_sensitive (ev_window, PAGE_SELECTOR_ACTION, document!=NULL);
+	set_action_sensitive (ev_window, PAGE_SELECTOR_ACTION, has_pages);
 }
 
 static void
@@ -295,6 +310,7 @@ ev_window_cmd_view_best_fit (GtkAction *action, EvWindow *ev_window)
 	} else {
 		ev_window_set_sizing_mode (ev_window, EV_SIZING_FREE);
 	}
+	update_action_sensitivity (ev_window);
 }
 
 static void
@@ -305,6 +321,7 @@ ev_window_cmd_view_page_width (GtkAction *action, EvWindow *ev_window)
 	} else {
 		ev_window_set_sizing_mode (ev_window, EV_SIZING_FREE);
 	}
+	update_action_sensitivity (ev_window);
 }
 
 static void
@@ -347,13 +364,10 @@ update_sizing_buttons (EvWindow *window)
 void
 ev_window_open_page (EvWindow *ev_window, int page)
 {
-	ev_view_set_page (EV_VIEW (ev_window->priv->view), page);
-}
-
-void
-ev_window_open_link (EvWindow *ev_window, EvLink *link)
-{
-	ev_view_go_to_link (EV_VIEW (ev_window->priv->view), link);
+	if (ev_window->priv->page_cache) {
+	    page = CLAMP (page, 0, ev_page_cache_get_n_pages (ev_window->priv->page_cache));
+	    ev_page_cache_set_current_page (ev_window->priv->page_cache, page);
+	}
 }
 
 gboolean
@@ -381,44 +395,6 @@ unable_to_load (EvWindow   *ev_window,
 	gtk_widget_destroy (dialog);
 }
 
-/* Would be nice to have this in gdk-pixbuf */
-static gboolean
-mime_type_supported_by_gdk_pixbuf (const gchar *mime_type)
-{
-	GSList *formats, *list;
-	gboolean retval = FALSE;
-
-	formats = gdk_pixbuf_get_formats ();
-
-	list = formats;
-	while (list) {
-		GdkPixbufFormat *format = list->data;
-		int i;
-		gchar **mime_types;
-
-		if (gdk_pixbuf_format_is_disabled (format))
-			continue;
-
-		mime_types = gdk_pixbuf_format_get_mime_types (format);
-
-		for (i = 0; mime_types[i] != NULL; i++) {
-			if (strcmp (mime_types[i], mime_type) == 0) {
-				retval = TRUE;
-				break;
-			}
-		}
-
-		if (retval)
-			break;
-
-		list = list->next;
-	}
-
-	g_slist_free (formats);
-
-	return retval;
-}
-
 static void
 update_window_title (EvDocument *document, GParamSpec *pspec, EvWindow *ev_window)
 {
@@ -427,9 +403,8 @@ update_window_title (EvDocument *document, GParamSpec *pspec, EvWindow *ev_windo
 	gboolean password_needed;
 
 	password_needed = (ev_window->priv->password_document != NULL);
-
-	if (document) {
-		doc_title = ev_document_get_title (document);
+	if (document && ev_window->priv->page_cache) {
+		doc_title = ev_page_cache_get_title (ev_window->priv->page_cache);
 
 		/* Make sure we get a valid title back */
 		if (doc_title) {
@@ -480,65 +455,67 @@ update_window_title (EvDocument *document, GParamSpec *pspec, EvWindow *ev_windo
 }
 
 static void
-update_total_pages (EvWindow *ev_window)
+find_changed_cb (EvDocument *document, int page, EvWindow *ev_window)
 {
-	GtkAction *action;
-	int pages;
-
-	pages = ev_document_get_n_pages (ev_window->priv->document);
-	action = gtk_action_group_get_action
-		(ev_window->priv->action_group, PAGE_SELECTOR_ACTION);
-	ev_page_action_set_total_pages (EV_PAGE_ACTION (action), pages);
-}
-
-/* This function assumes that ev_window just had ev_window->document set.
- */
-static gboolean
-document_supports_sidebar (EvDocument *document)
-{
-	return (EV_IS_DOCUMENT_THUMBNAILS (document) && EV_IS_DOCUMENT_LINKS (document));
+	update_action_sensitivity (ev_window);
 }
 
 static void
-hide_sidebar_and_actions (EvWindow *ev_window)
+page_changed_cb (EvPageCache *page_cache,
+		 gint         page,
+		 EvWindow    *ev_window)
 {
-	GtkAction *action;
-	/* Alsthough we update the hiddenness of the sidebar, we don't want to
-	 * store the value */
-	g_signal_handlers_disconnect_by_func (ev_window->priv->sidebar,
-					      ev_window_sidebar_visibility_changed_cb,
-					      ev_window);
-	gtk_widget_hide (ev_window->priv->sidebar);
-	action = gtk_action_group_get_action (ev_window->priv->action_group, "ViewSidebar");
-	gtk_action_set_sensitive (action, FALSE);
+	update_action_sensitivity (ev_window);
+}
 
+static void
+update_document_mode (EvWindow *window, EvDocumentMode mode)
+{
+	if (mode == EV_DOCUMENT_MODE_FULL_SCREEN) {
+		gtk_window_fullscreen (GTK_WINDOW (window));
+	}
 }
 
 static void
 ev_window_setup_document (EvWindow *ev_window)
 {
+	EvDocumentInfo *info;
 	EvDocument *document;
 	EvView *view = EV_VIEW (ev_window->priv->view);
 	EvSidebar *sidebar = EV_SIDEBAR (ev_window->priv->sidebar);
+	GtkAction *action;
 
 	document = ev_window->priv->document;
+	ev_window->priv->page_cache = ev_document_get_page_cache (ev_window->priv->document);
+	g_signal_connect (ev_window->priv->page_cache, "page-changed", G_CALLBACK (page_changed_cb), ev_window);
 
 	g_signal_connect_object (G_OBJECT (document),
 				 "notify::title",
 				 G_CALLBACK (update_window_title),
 				 ev_window, 0);
+	if (EV_IS_DOCUMENT_FIND (document)) {
+		g_signal_connect_object (G_OBJECT (document),
+				         "find_changed",
+				         G_CALLBACK (find_changed_cb),	
+				         ev_window, 0);
+	}
 
 	ev_window_set_page_mode (ev_window, PAGE_MODE_SINGLE_PAGE);
 
-	if (document_supports_sidebar (document)) 
-		ev_sidebar_set_document (sidebar, document);
-	else
-		hide_sidebar_and_actions (ev_window);
-	ev_view_set_document (view, document);
+	ev_sidebar_set_document (sidebar, document);
+
+	if (ev_page_cache_get_n_pages (ev_window->priv->page_cache) > 0) {
+		ev_view_set_document (view, document);
+	}
 
 	update_window_title (document, NULL, ev_window);
-	update_total_pages (ev_window);
+	action = gtk_action_group_get_action (ev_window->priv->action_group, PAGE_SELECTOR_ACTION);
+	ev_page_action_set_document (EV_PAGE_ACTION (action), document);
 	update_action_sensitivity (ev_window);
+
+	info = ev_document_get_info (document);
+	update_document_mode (ev_window, info->mode);
+	g_free (info);
 }
 
 static void
@@ -553,9 +530,12 @@ password_dialog_response (GtkWidget *password_dialog,
 		gchar *uri;
 
 		password = ev_password_dialog_get_password (password_dialog);
-		if (password)
+		if (password) {
+			ev_document_doc_mutex_lock ();
 			ev_document_security_set_password (EV_DOCUMENT_SECURITY (ev_window->priv->password_document),
 							   password);
+			ev_document_doc_mutex_unlock ();
+		}
 		g_free (password);
 
 		document = ev_window->priv->password_document;
@@ -649,6 +629,8 @@ start_loading_document (EvWindow   *ev_window,
 			g_object_unref (ev_window->priv->document);
 		ev_window->priv->document = g_object_ref (document);
 		ev_window_setup_document (ev_window);
+		
+		ev_window_add_recent (ev_window, uri);
 
 		return TRUE;
 	}
@@ -658,15 +640,18 @@ start_loading_document (EvWindow   *ev_window,
 
 	if (error->domain == EV_DOCUMENT_ERROR &&
 	    error->code == EV_DOCUMENT_ERROR_ENCRYPTED) {
-		char *file_name;
+		gchar *base_name, *file_name;
 
 		ev_window->priv->password_document = g_object_ref (document);
 		ev_window->priv->password_uri = g_strdup (uri);
 
-		file_name = g_path_get_basename (uri);
+		base_name = g_path_get_basename (uri);
+		file_name = gnome_vfs_unescape_string_for_display (base_name);
+
 		ev_password_view_set_file_name (EV_PASSWORD_VIEW (ev_window->priv->password_view),
 						file_name);
 		g_free (file_name);
+		g_free (base_name);
 		ev_window_set_page_mode (ev_window, PAGE_MODE_PASSWORD);
 
 		ev_window_popup_password_dialog (ev_window);
@@ -678,31 +663,6 @@ start_loading_document (EvWindow   *ev_window,
 	return FALSE;
 }
 
-static gboolean
-is_file_supported (const gchar *mime_type)
-{
-	static char *supported_types [] = {
-		"application/pdf",
-		"application/postscript",
-		"application/x-gzpostscript",
-		"image/x-eps",
-		NULL
-	};
-	gint   i;
-	
-	g_return_val_if_fail (mime_type != NULL, FALSE);
-
-	if (mime_type_supported_by_gdk_pixbuf (mime_type))
-		return TRUE;
-	
-	for (i = 0; supported_types[i] != NULL; i++) {
-		if (g_ascii_strcasecmp (mime_type, supported_types[i]) == 0)
-			return TRUE;
-	}
-	
-	return FALSE;
-}
-
 void
 ev_window_open (EvWindow *ev_window, const char *uri)
 {
@@ -716,14 +676,13 @@ ev_window_open (EvWindow *ev_window, const char *uri)
 
 	if (mime_type == NULL)
 		document = NULL;
-	else if (!strcmp (mime_type, "application/pdf"))
-		document = g_object_new (PDF_TYPE_DOCUMENT, NULL);
-	else if (!strcmp (mime_type, "application/postscript") ||
-		 !strcmp (mime_type, "application/x-gzpostscript") ||
-		 !strcmp (mime_type, "image/x-eps"))
-		document = g_object_new (PS_TYPE_DOCUMENT, NULL);
-	else if (mime_type_supported_by_gdk_pixbuf (mime_type))
-		document = g_object_new (PIXBUF_TYPE_DOCUMENT, NULL);
+	else {
+		GType document_type = ev_document_type_lookup (mime_type);
+
+		if (document_type!=G_TYPE_INVALID) {
+			document = g_object_new (document_type, NULL);
+		}
+	}
 
 	if (document) {
 		start_loading_document (ev_window, document, uri);
@@ -756,7 +715,7 @@ ev_window_open_uri_list (EvWindow *ev_window, GList *uri_list)
 		uri = gnome_vfs_uri_to_string (list->data, GNOME_VFS_URI_HIDE_NONE);
 		mime_type = gnome_vfs_get_mime_type (uri);
 		
-		if (is_file_supported (mime_type)) {
+		if (ev_document_type_lookup (mime_type)!=G_TYPE_INVALID) {
 			if (ev_window_is_empty (EV_WINDOW (ev_window))) {
 				ev_window_open (ev_window, uri);
 				
@@ -784,6 +743,69 @@ ev_window_cmd_file_open (GtkAction *action, EvWindow *ev_window)
 	ev_application_open (EV_APP, NULL);
 }
 
+static void
+ev_window_cmd_recent_file_activate (EggRecentViewGtk *view, EggRecentItem *item,
+        	                    EvWindow *ev_window)
+{
+	char *uri;
+	GtkWidget *window;
+
+	uri = egg_recent_item_get_uri (item);
+
+	window = GTK_WIDGET (ev_application_get_empty_window (EV_APP));
+	gtk_widget_show (window);
+	ev_window_open (EV_WINDOW (window), uri);
+	
+	g_free (uri);
+}
+
+static void
+ev_window_add_recent (EvWindow *window, const char *filename)
+{
+	EggRecentItem *item;
+
+	if (strstr (filename, "file:///") == NULL)
+		return;
+
+	item = egg_recent_item_new_from_uri (filename);
+	egg_recent_item_add_group (item, "Evince");
+	egg_recent_model_add_full (window->priv->recent_model, item);
+}
+
+static void
+ev_window_setup_recent (EvWindow *ev_window)
+{
+	GtkWidget *menu_item;
+	GtkWidget *menu;
+
+	menu_item = gtk_ui_manager_get_widget (ev_window->priv->ui_manager, "/MainMenu/FileMenu");
+	menu = gtk_menu_item_get_submenu (GTK_MENU_ITEM (menu_item));
+	menu_item = gtk_ui_manager_get_widget (ev_window->priv->ui_manager, "/MainMenu/FileMenu/RecentFilesMenu");
+
+	g_return_if_fail (menu != NULL);
+	g_return_if_fail (menu_item != NULL);
+
+	/* it would be better if we just filtered by mime-type, but there
+	 * doesn't seem to be an easy way to figure out which mime-types we
+	 * can handle */
+	ev_window->priv->recent_model = egg_recent_model_new (EGG_RECENT_MODEL_SORT_MRU);
+
+	ev_window->priv->recent_view = egg_recent_view_gtk_new (menu, menu_item);
+	egg_recent_view_gtk_show_icons (EGG_RECENT_VIEW_GTK (ev_window->priv->recent_view), FALSE);
+	egg_recent_model_set_limit (ev_window->priv->recent_model, 5);
+
+	egg_recent_view_set_model (EGG_RECENT_VIEW (ev_window->priv->recent_view),
+				   ev_window->priv->recent_model);
+
+	egg_recent_model_set_filter_groups (ev_window->priv->recent_model,
+    	    	    			    "Evince", NULL);
+
+	egg_recent_view_gtk_set_trailing_sep (ev_window->priv->recent_view, TRUE);
+	
+	g_signal_connect (ev_window->priv->recent_view, "activate",
+			G_CALLBACK (ev_window_cmd_recent_file_activate), ev_window);
+}
+
 /* FIXME
 static gboolean
 overwrite_existing_file (GtkWindow *window, const gchar *file_name)
@@ -856,6 +878,7 @@ ev_window_cmd_save_as (GtkAction *action, EvWindow *ev_window)
 	GtkWidget *fc;
 	GtkFileFilter *pdf_filter, *all_filter;
 	gchar *uri = NULL;
+	gboolean success;
 
 	fc = gtk_file_chooser_dialog_new (
 		_("Save a Copy"),
@@ -888,8 +911,12 @@ ev_window_cmd_save_as (GtkAction *action, EvWindow *ev_window)
 		    !overwrite_existing_file (GTK_WINDOW (fc), uri))
 				continue;
 */
+		
+		ev_document_doc_mutex_lock ();
+		success = ev_document_save (ev_window->priv->document, uri, NULL);
+		ev_document_doc_mutex_unlock ();
 
-		if (ev_document_save (ev_window->priv->document, uri, NULL))
+		if (success)
 			break;
 		else
 			save_error_dialog (GTK_WINDOW (fc), uri);
@@ -936,7 +963,7 @@ ev_window_print (EvWindow *ev_window)
 	config = gnome_print_config_default ();
 	job = gnome_print_job_new (config);
 
-	print_dialog = gnome_print_dialog_new (job, _("Print"),
+	print_dialog = gnome_print_dialog_new (job, (guchar *) _("Print"),
 					       (GNOME_PRINT_DIALOG_RANGE |
 						GNOME_PRINT_DIALOG_COPIES));
 	gtk_dialog_set_response_sensitive (GTK_DIALOG (print_dialog),
@@ -964,7 +991,7 @@ ev_window_print (EvWindow *ev_window)
 				GTK_MESSAGE_DIALOG (dialog),
 				_("You were trying to print to a printer using the \"%s\" driver. This program requires a PostScript printer driver."),
 				gnome_print_config_get (
-					config, "Settings.Engine.Backend.Driver"));
+					config, (guchar *)"Settings.Engine.Backend.Driver"));
 			gtk_dialog_run (GTK_DIALOG (dialog));
 			gtk_widget_destroy (dialog);
 
@@ -1041,10 +1068,18 @@ ev_window_cmd_edit_find (GtkAction *action, EvWindow *ev_window)
 	} else {
 		update_chrome_flag (ev_window, EV_CHROME_FINDBAR, NULL, TRUE);
 
-		egg_find_bar_grab_focus (EGG_FIND_BAR (ev_window->priv->find_bar));
+		gtk_widget_grab_focus (ev_window->priv->find_bar);
 	}
 }
 
+static void
+ev_window_cmd_edit_find_next (GtkAction *action, EvWindow *ev_window)
+{
+        g_return_if_fail (EV_IS_WINDOW (ev_window));
+
+	ev_view_find_next (EV_VIEW (ev_window->priv->view));
+}
+
 static void
 ev_window_cmd_edit_copy (GtkAction *action, EvWindow *ev_window)
 {
@@ -1301,6 +1336,8 @@ ev_window_fullscreen (EvWindow *window)
 static void
 ev_window_unfullscreen (EvWindow *window)
 {
+	EvView *view = EV_VIEW (window->priv->view);
+
 	window->priv->fullscreen_mode = FALSE;
 
 	g_object_set (G_OBJECT (window->priv->scrolled_window),
@@ -1309,48 +1346,60 @@ ev_window_unfullscreen (EvWindow *window)
 
 	fullscreen_clear_timeout (window);
 
-	g_signal_handlers_disconnect_by_func (window->priv->view,
+	g_signal_handlers_disconnect_by_func (view,
 					      (gpointer) fullscreen_motion_notify_cb,
 					      window);
 
-//	destroy_fullscreen_popup (window);
-
+	ev_view_set_show_border (view, TRUE);
+	ev_view_set_spacing (view, VIEW_SPACING);	
 	update_chrome_visibility (window);
 }
 
 static void
-ev_window_cmd_view_fullscreen (GtkAction *action, EvWindow *ev_window)
+ev_window_cmd_view_fullscreen (GtkAction *action, EvWindow *window)
 {
+	EvView *view;
 	gboolean fullscreen;
 
-        g_return_if_fail (EV_IS_WINDOW (ev_window));
+        g_return_if_fail (EV_IS_WINDOW (window));
 
+	view = EV_VIEW (window->priv->view);
 	fullscreen = gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action));
 
 	if (fullscreen) {
-		gtk_window_fullscreen (GTK_WINDOW (ev_window));
+		gtk_window_fullscreen (GTK_WINDOW (window));
 	} else {
-		gtk_window_unfullscreen (GTK_WINDOW (ev_window));
+		gtk_window_unfullscreen (GTK_WINDOW (window));
 	}
+
+	ev_view_set_show_border (view, FALSE);
+	ev_view_set_spacing (view, 0);	
 }
 
 static gboolean
-ev_window_state_event_cb (GtkWidget *widget, GdkEventWindowState *event, EvWindow *window)
+ev_window_state_event (GtkWidget *widget, GdkEventWindowState *event)
 {
-	if (event->changed_mask & GDK_WINDOW_STATE_FULLSCREEN)
-	{
+	EvWindow *window = EV_WINDOW (widget);
+
+	if (event->changed_mask & GDK_WINDOW_STATE_MAXIMIZED) {
+		gboolean show;
+
+		show = (event->new_window_state & GDK_WINDOW_STATE_MAXIMIZED) == 0;
+
+		gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR (window->priv->statusbar),
+						   show);
+	}
+
+	if (event->changed_mask & GDK_WINDOW_STATE_FULLSCREEN) {
 		GtkActionGroup *action_group;
 		GtkAction *action;
 		gboolean fullscreen;
 
 		fullscreen = event->new_window_state & GDK_WINDOW_STATE_FULLSCREEN;
 
-		if (fullscreen)
-		{
+		if (fullscreen)	{
 			ev_window_fullscreen (window);
-		}
-		else
-		{
+		} else {
 			ev_window_unfullscreen (window);
 		}
 
@@ -1362,7 +1411,6 @@ ev_window_state_event_cb (GtkWidget *widget, GdkEventWindowState *event, EvWindo
 		gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action), fullscreen);
 		g_signal_handlers_unblock_by_func
 			(action, G_CALLBACK (ev_window_cmd_view_fullscreen), window);
-
 	}
 
 	return FALSE;
@@ -1438,8 +1486,8 @@ ev_window_cmd_view_zoom_in (GtkAction *action, EvWindow *ev_window)
         g_return_if_fail (EV_IS_WINDOW (ev_window));
 
 	ev_window_set_sizing_mode (ev_window, EV_SIZING_FREE);
-
 	ev_view_zoom_in (EV_VIEW (ev_window->priv->view));
+	update_action_sensitivity (ev_window);
 }
 
 static void
@@ -1448,8 +1496,8 @@ ev_window_cmd_view_zoom_out (GtkAction *action, EvWindow *ev_window)
         g_return_if_fail (EV_IS_WINDOW (ev_window));
 
 	ev_window_set_sizing_mode (ev_window, EV_SIZING_FREE);
-
 	ev_view_zoom_out (EV_VIEW (ev_window->priv->view));
+	update_action_sensitivity (ev_window);
 }
 
 static void
@@ -1457,7 +1505,8 @@ ev_window_cmd_view_normal_size (GtkAction *action, EvWindow *ev_window)
 {
         g_return_if_fail (EV_IS_WINDOW (ev_window));
 
-	ev_view_set_size (EV_VIEW (ev_window->priv->view), -1, -1);
+	ev_view_zoom_normal (EV_VIEW (ev_window->priv->view));
+	update_action_sensitivity (ev_window);
 }
 
 static void
@@ -1465,8 +1514,7 @@ ev_window_cmd_go_previous_page (GtkAction *action, EvWindow *ev_window)
 {
         g_return_if_fail (EV_IS_WINDOW (ev_window));
 
-	ev_view_set_page (EV_VIEW (ev_window->priv->view),
-			  ev_view_get_page (EV_VIEW (ev_window->priv->view)) - 1);
+	ev_page_cache_prev_page (ev_window->priv->page_cache);
 }
 
 static void
@@ -1474,8 +1522,7 @@ ev_window_cmd_go_next_page (GtkAction *action, EvWindow *ev_window)
 {
         g_return_if_fail (EV_IS_WINDOW (ev_window));
 
-	ev_view_set_page (EV_VIEW (ev_window->priv->view),
-			  ev_view_get_page (EV_VIEW (ev_window->priv->view)) + 1);
+	ev_page_cache_next_page (ev_window->priv->page_cache);
 }
 
 static void
@@ -1483,15 +1530,18 @@ ev_window_cmd_go_first_page (GtkAction *action, EvWindow *ev_window)
 {
         g_return_if_fail (EV_IS_WINDOW (ev_window));
 
-	ev_view_set_page (EV_VIEW (ev_window->priv->view), 1);
+	ev_page_cache_set_current_page (ev_window->priv->page_cache, 0);
 }
 
 static void
 ev_window_cmd_go_last_page (GtkAction *action, EvWindow *ev_window)
 {
+	int n_pages;
+
         g_return_if_fail (EV_IS_WINDOW (ev_window));
 
-	ev_view_set_page (EV_VIEW (ev_window->priv->view), G_MAXINT);
+	n_pages = ev_page_cache_get_n_pages (ev_window->priv->page_cache);
+	ev_page_cache_set_current_page (ev_window->priv->page_cache, n_pages - 1);
 }
 
 static void
@@ -1502,7 +1552,7 @@ ev_window_cmd_view_reload (GtkAction *action, EvWindow *ev_window)
 
 	g_return_if_fail (EV_IS_WINDOW (ev_window));
 
-	page = ev_document_get_page (ev_window->priv->document);
+	page = ev_page_cache_get_current_page (ev_window->priv->page_cache);
 	uri = g_strdup (ev_window->priv->uri);
 
 	ev_window_open (ev_window, uri);
@@ -1514,9 +1564,16 @@ ev_window_cmd_view_reload (GtkAction *action, EvWindow *ev_window)
 static void
 ev_window_cmd_help_contents (GtkAction *action, EvWindow *ev_window)
 {
+	GError *error = NULL;
+
         g_return_if_fail (EV_IS_WINDOW (ev_window));
 
-        /* FIXME */
+	gnome_help_display ("evince.xml", NULL, &error);
+
+	if(error != NULL) {
+		g_warning (error->message);
+		g_error_free (error);
+	}
 }
 
 static void
@@ -1525,6 +1582,19 @@ ev_window_cmd_leave_fullscreen (GtkAction *action, EvWindow *window)
 	gtk_window_unfullscreen (GTK_WINDOW (window));
 }
 
+static void
+ev_window_cmd_escape (GtkAction *action, EvWindow *window)
+{
+	GtkWidget *widget;
+
+	widget = gtk_window_get_focus (GTK_WINDOW (window));
+	if (widget && gtk_widget_get_ancestor (widget, EGG_TYPE_FIND_BAR)) {
+		update_chrome_flag (window, EV_CHROME_FINDBAR, NULL, FALSE);
+	} else {
+		gtk_window_unfullscreen (GTK_WINDOW (window));
+	}
+}
+
 static void
 update_view_size (EvWindow *window)
 {
@@ -1535,9 +1605,12 @@ update_view_size (EvWindow *window)
 	width = window->priv->scrolled_window->allocation.width;
 	height = window->priv->scrolled_window->allocation.height;
 
-	/* the scrolled window has a GTK_SHADOW_IN */
-	width -= 2 * window->priv->view->style->xthickness;
-	height -= 2 * window->priv->view->style->ythickness;
+	if (gtk_scrolled_window_get_shadow_type
+		(GTK_SCROLLED_WINDOW (window->priv->scrolled_window)) == GTK_SHADOW_IN)
+	{
+		width -= 2 * window->priv->view->style->xthickness;
+		height -= 2 * window->priv->view->style->ythickness;
+	}
 
 	if (window->priv->sizing_mode == EV_SIZING_BEST_FIT) {
 		ev_view_set_size (EV_VIEW (window->priv->view),
@@ -1766,32 +1839,6 @@ disconnect_proxy_cb (GtkUIManager *ui_manager, GtkAction *action,
 	}
 }
 
-static void
-update_current_page (EvWindow *ev_window,
-		     EvView   *view)
-{
-	int page;
-	GtkAction *action;
-	EvSidebarThumbnails *thumbs;
-
-	thumbs = EV_SIDEBAR_THUMBNAILS (ev_window->priv->thumbs_sidebar);
-	ev_sidebar_thumbnails_select_page (thumbs, ev_view_get_page (view));
-
-	action = gtk_action_group_get_action
-		(ev_window->priv->action_group, PAGE_SELECTOR_ACTION);
-
-	page = ev_view_get_page (EV_VIEW (ev_window->priv->view));
-	ev_page_action_set_current_page (EV_PAGE_ACTION (action), page);
-}
-
-static void
-view_page_changed_cb (EvView   *view,
-		      EvWindow *ev_window)
-{
-	update_current_page (ev_window, view);
-	update_action_sensitivity (ev_window);
-}
-
 static void
 view_status_changed_cb (EvView     *view,
 			GParamSpec *pspec,
@@ -1884,10 +1931,18 @@ find_bar_search_changed_cb (EggFindBar *find_bar,
 
 	if (ev_window->priv->document &&
 	    EV_IS_DOCUMENT_FIND (ev_window->priv->document)) {
-		if (visible && search_string) {
-			ev_document_find_begin (EV_DOCUMENT_FIND (ev_window->priv->document), search_string, case_sensitive);
+		if (visible && search_string && search_string[0]) {
+			ev_document_doc_mutex_lock ();
+			ev_document_find_begin (EV_DOCUMENT_FIND (ev_window->priv->document), 
+						ev_page_cache_get_current_page (ev_window->priv->page_cache),
+						search_string,
+						case_sensitive);
+			ev_document_doc_mutex_unlock ();
 		} else {
+			ev_document_doc_mutex_lock ();
 			ev_document_find_cancel (EV_DOCUMENT_FIND (ev_window->priv->document));
+			ev_document_doc_mutex_unlock ();
+
 			egg_find_bar_set_status_text (EGG_FIND_BAR (ev_window->priv->find_bar),
 						      NULL);
 			gtk_widget_queue_draw (GTK_WIDGET (ev_window->priv->view));
@@ -1936,6 +1991,14 @@ ev_window_dispose (GObject *object)
 		priv->password_uri = NULL;
 	}
 
+	if (priv->find_bar) {
+		g_signal_handlers_disconnect_by_func
+			(window->priv->find_bar,
+			 G_CALLBACK (find_bar_close_cb),
+			 window);
+		priv->find_bar = NULL;
+	}
+
 	destroy_fullscreen_popup (window);
 
 	G_OBJECT_CLASS (ev_window_parent_class)->dispose (object);
@@ -1949,6 +2012,7 @@ ev_window_class_init (EvWindowClass *ev_window_class)
 
 	g_object_class->dispose = ev_window_dispose;
 
+	widget_class->window_state_event = ev_window_state_event;
 	widget_class->focus_in_event = ev_window_focus_in_event;
 	widget_class->focus_out_event = ev_window_focus_out_event;
 
@@ -1956,7 +2020,7 @@ ev_window_class_init (EvWindowClass *ev_window_class)
 }
 
 /* Normal items */
-static GtkActionEntry entries[] = {
+static const GtkActionEntry entries[] = {
 	{ "File", NULL, N_("_File") },
         { "Edit", NULL, N_("_Edit") },
 	{ "View", NULL, N_("_View") },
@@ -1987,6 +2051,9 @@ static GtkActionEntry entries[] = {
         { "EditFind", GTK_STOCK_FIND, NULL, "<control>F",
           N_("Find a word or phrase in the document"),
           G_CALLBACK (ev_window_cmd_edit_find) },
+	{ "EditFindNext", NULL, N_("Find Ne_xt"), "<control>G",
+	  N_("Find next occurrence of the word or phrase"),
+	  G_CALLBACK (ev_window_cmd_edit_find_next) },
 
         /* View menu */
         { "ViewZoomIn", GTK_STOCK_ZOOM_IN, NULL, "<control>plus",
@@ -2017,7 +2084,7 @@ static GtkActionEntry entries[] = {
           G_CALLBACK (ev_window_cmd_go_last_page) },
 
 	/* Help menu */
-	{ "HelpContents", GTK_STOCK_HELP, N_("_Contents"), NULL,
+	{ "HelpContents", GTK_STOCK_HELP, N_("_Contents"), "F1",
 	  N_("Display help for the viewer application"),
 	  G_CALLBACK (ev_window_cmd_help_contents) },
 
@@ -2026,13 +2093,16 @@ static GtkActionEntry entries[] = {
 	  G_CALLBACK (ev_window_cmd_help_about) },
 
 	/* Toolbar-only */
-	{ "LeaveFullscreen", EV_STOCK_LEAVE_FULLSCREEN, N_("Leave Fullscreen"), "Escape",
+	{ "LeaveFullscreen", EV_STOCK_LEAVE_FULLSCREEN, N_("Leave Fullscreen"), NULL,
 	  N_("Leave fullscreen mode"),
-	  G_CALLBACK (ev_window_cmd_leave_fullscreen) }
+	  G_CALLBACK (ev_window_cmd_leave_fullscreen) },
+
+	{ "Escape", NULL, N_("Selection Caret"), "Escape", "",
+	  G_CALLBACK (ev_window_cmd_escape) }
 };
 
 /* Toggle items */
-static GtkToggleActionEntry toggle_entries[] = {
+static const GtkToggleActionEntry toggle_entries[] = {
 	/* View Menu */
 	{ "ViewToolbar", NULL, N_("_Toolbar"), "<shift><control>T",
 	  N_("Show or hide the toolbar"),
@@ -2054,7 +2124,7 @@ static GtkToggleActionEntry toggle_entries[] = {
           G_CALLBACK (ev_window_cmd_view_page_width) },
 };
 
-static GtkRadioActionEntry page_view_entries[] = {
+static const GtkRadioActionEntry page_view_entries[] = {
 	{ "SinglePage", GTK_STOCK_DND, N_("Single"), NULL,
 	  N_("Show the document one page at a time"),
 	  PAGE_MODE_SINGLE_PAGE },
@@ -2063,16 +2133,6 @@ static GtkRadioActionEntry page_view_entries[] = {
 	  PAGE_MODE_CONTINUOUS_PAGE }
 };
 
-static void
-goto_page_cb (GtkAction *action, int page_number, EvWindow *ev_window)
-{
-	EvView *view = EV_VIEW (ev_window->priv->view);
-
-	if (ev_view_get_page (view) != page_number) {
-		ev_view_set_page (view, page_number);
-	}
-}
-
 static void
 drag_data_received_cb (GtkWidget *widget, GdkDragContext *context,
 		       gint x, gint y, GtkSelectionData *selection_data,
@@ -2101,8 +2161,6 @@ register_custom_actions (EvWindow *window, GtkActionGroup *group)
 			       "label", _("Page"),
 			       "tooltip", _("Select Page"),
 			       NULL);
-	g_signal_connect (action, "goto_page",
-			  G_CALLBACK (goto_page_cb), window);
 	gtk_action_group_add_action (group, action);
 	g_object_unref (action);
 }
@@ -2116,10 +2174,27 @@ set_action_properties (GtkActionGroup *action_group)
 	/*translators: this is the label for toolbar button*/
 	g_object_set (action, "short_label", _("Previous"), NULL);
 	g_object_set (action, "is-important", TRUE, NULL);
+
 	action = gtk_action_group_get_action (action_group, "GoNextPage");
-	/*translators: this is the label for toolbar button*/
 	g_object_set (action, "is-important", TRUE, NULL);
+	/*translators: this is the label for toolbar button*/
 	g_object_set (action, "short_label", _("Next"), NULL);
+
+	action = gtk_action_group_get_action (action_group, "ViewZoomIn");
+	/*translators: this is the label for toolbar button*/
+	g_object_set (action, "short_label", _("Zoom In"), NULL);
+	action = gtk_action_group_get_action (action_group, "ViewZoomIn");
+
+	action = gtk_action_group_get_action (action_group, "ViewZoomOut");
+	/*translators: this is the label for toolbar button*/
+	g_object_set (action, "short_label", _("Zoom Out"), NULL);
+	action = gtk_action_group_get_action (action_group, "ViewZoomIn");
+
+	action = gtk_action_group_get_action (action_group, "ViewBestFit");
+	/*translators: this is the label for toolbar button*/
+	g_object_set (action, "short_label", _("Best Fit"), NULL);
+	action = gtk_action_group_get_action (action_group, "ViewZoomIn");
+
 	action = gtk_action_group_get_action (action_group, "ViewPageWidth");
 	/*translators: this is the label for toolbar button*/
 	g_object_set (action, "short_label", _("Fit Width"), NULL);
@@ -2199,6 +2274,22 @@ load_chrome (void)
 	return chrome;
 }
 
+static void
+sidebar_widget_model_set (EvSidebarLinks *ev_sidebar_links,
+			  GParamSpec     *pspec,
+			  EvWindow       *ev_window)
+{
+	GtkTreeModel *model;
+	GtkAction *action;
+
+	g_object_get (G_OBJECT (ev_sidebar_links),
+		      "model", &model,
+		      NULL);
+
+	action = gtk_action_group_get_action (ev_window->priv->action_group, PAGE_SELECTOR_ACTION);
+	ev_page_action_set_model (EV_PAGE_ACTION (action), model);
+}
+
 static void
 ev_window_init (EvWindow *ev_window)
 {
@@ -2281,7 +2372,6 @@ ev_window_init (EvWindow *ev_window)
 	ev_window->priv->fullscreen_toolbar =
 		gtk_ui_manager_get_widget (ev_window->priv->ui_manager, "/LeaveFullscreenToolbar");
 	gtk_toolbar_set_show_arrow (GTK_TOOLBAR (ev_window->priv->fullscreen_toolbar), TRUE);
-	gtk_toolbar_set_style (GTK_TOOLBAR (ev_window->priv->fullscreen_toolbar), GTK_TOOLBAR_BOTH_HORIZ);
 
 	/* Add the main area */
 	ev_window->priv->hpaned = gtk_hpaned_new ();
@@ -2312,18 +2402,18 @@ ev_window_init (EvWindow *ev_window)
 
 	/* Stub sidebar, for now */
 	sidebar_widget = ev_sidebar_links_new ();
+	g_signal_connect (sidebar_widget,
+			  "notify::model",
+			  G_CALLBACK (sidebar_widget_model_set),
+			  ev_window);
 	gtk_widget_show (sidebar_widget);
 	ev_sidebar_add_page (EV_SIDEBAR (ev_window->priv->sidebar),
-			     "index",
-			     _("Index"),
 			     sidebar_widget);
 
-	ev_window->priv->thumbs_sidebar = ev_sidebar_thumbnails_new ();
-	gtk_widget_show (ev_window->priv->thumbs_sidebar);
+	sidebar_widget = ev_sidebar_thumbnails_new ();
+	gtk_widget_show (sidebar_widget);
 	ev_sidebar_add_page (EV_SIDEBAR (ev_window->priv->sidebar),
-			     "thumbnails",
-			     _("Thumbnails"),
-			     ev_window->priv->thumbs_sidebar);
+			     sidebar_widget);
 
 	ev_window->priv->scrolled_window =
 		GTK_WIDGET (g_object_new (GTK_TYPE_SCROLLED_WINDOW,
@@ -2335,7 +2425,7 @@ ev_window_init (EvWindow *ev_window)
 			ev_window->priv->scrolled_window);
 
 	ev_window->priv->view = ev_view_new ();
-	//ev_window->priv->page_view = ev_page_view_new ();
+	ev_view_set_spacing (EV_VIEW (ev_window->priv->view), VIEW_SPACING);	
 	ev_window->priv->password_view = ev_password_view_new ();
 	g_signal_connect_swapped (ev_window->priv->password_view,
 				  "unlock",
@@ -2352,10 +2442,6 @@ ev_window_init (EvWindow *ev_window)
 
 	gtk_container_add (GTK_CONTAINER (ev_window->priv->scrolled_window),
 			   ev_window->priv->view);
-	g_signal_connect (ev_window->priv->view,
-			  "page-changed",
-			  G_CALLBACK (view_page_changed_cb),
-			  ev_window);
 	g_signal_connect (ev_window->priv->view,
 			  "notify::find-status",
 			  G_CALLBACK (view_find_status_changed_cb),
@@ -2379,6 +2465,7 @@ ev_window_init (EvWindow *ev_window)
 			  ev_window->priv->find_bar,
 			  FALSE, TRUE, 0);
 
+	ev_window_setup_recent (ev_window);
 	ev_window->priv->chrome = load_chrome ();
 	set_chrome_actions (ev_window);
 	update_chrome_visibility (ev_window);
@@ -2415,10 +2502,6 @@ ev_window_init (EvWindow *ev_window)
 			  G_CALLBACK (find_bar_search_changed_cb),
 			  ev_window);
 
-	g_signal_connect (ev_window, "window-state-event",
-			  G_CALLBACK (ev_window_state_event_cb),
-			  ev_window);
-
 	/* Give focus to the scrolled window */
 	gtk_widget_grab_focus (ev_window->priv->scrolled_window);
 
