@@ -48,17 +48,11 @@ struct _EvWindow {
 
 struct _EvWindowClass {
 	GtkWindowClass		base_class;
-
-	/* signals */
-	void (*signal)		(EvWindow	*self,
-				 const char	*string);
 };
 
 GType		ev_window_get_type	(void);
 void		ev_window_open		(EvWindow       *ev_window,
 					 const char     *uri);
-void		ev_window_open_link	(EvWindow       *ev_window,
-					 EvLink         *link);
 void		ev_window_open_page	(EvWindow       *ev_window,
 					 int             page);
 gboolean	ev_window_is_empty	(const EvWindow *ev_window);
